@@ -24,7 +24,8 @@
 import os
 import tempfile
 import warnings
-from typing import Callable, Optional
+from collections.abc import Sequence
+from typing import Callable, Optional, TypedDict
 
 from .index import Index, build_index_from_tree
 from .object_store import BaseObjectStore
@@ -32,6 +33,19 @@ from .objects import Commit, Tag, Tree
 from .refs import RefsContainer
 
 
+class CommitData(TypedDict, total=False):
+    """TypedDict for commit data fields."""
+
+    author: bytes
+    author_time: int
+    author_timezone: int
+    committer: bytes
+    commit_time: int
+    commit_timezone: int
+    message: bytes
+    encoding: bytes
+
+
 class CommitFilter:
     """Filter for rewriting commits during filter-branch operations."""
 
@@ -39,13 +53,13 @@ class CommitFilter:
         self,
         object_store: BaseObjectStore,
         *,
-        filter_fn: Optional[Callable[[Commit], Optional[dict[str, bytes]]]] = None,
+        filter_fn: Optional[Callable[[Commit], Optional[CommitData]]] = None,
         filter_author: Optional[Callable[[bytes], Optional[bytes]]] = None,
         filter_committer: Optional[Callable[[bytes], Optional[bytes]]] = None,
         filter_message: Optional[Callable[[bytes], Optional[bytes]]] = None,
         tree_filter: Optional[Callable[[bytes, str], Optional[bytes]]] = None,
         index_filter: Optional[Callable[[bytes, str], Optional[bytes]]] = None,
-        parent_filter: Optional[Callable[[list[bytes]], list[bytes]]] = None,
+        parent_filter: Optional[Callable[[Sequence[bytes]], list[bytes]]] = None,
         commit_filter: Optional[Callable[[Commit, bytes], Optional[bytes]]] = None,
         subdirectory_filter: Optional[bytes] = None,
         prune_empty: bool = False,
@@ -122,6 +136,7 @@ class CommitFilter:
             for entry in current_tree.items():
                 if entry.path == part:
                     try:
+                        assert entry.sha is not None
                         obj = self.object_store[entry.sha]
                         if isinstance(obj, Tree):
                             current_tree = obj
@@ -267,7 +282,7 @@ class CommitFilter:
                 return new_parents[0]
 
         # Apply filters
-        new_data = {}
+        new_data: CommitData = {}
 
         # Custom filter function takes precedence
         if self.filter_fn:
@@ -363,7 +378,7 @@ class CommitFilter:
 def filter_refs(
     refs: RefsContainer,
     object_store: BaseObjectStore,
-    ref_names: list[bytes],
+    ref_names: Sequence[bytes],
     commit_filter: CommitFilter,
     *,
     keep_original: bool = True,
